// Copyright 2019 Uber Technologies, Inc. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// ============================================================================

#ifndef HOROVOD_GLOO_MEMORY_STORE_H
#define HOROVOD_GLOO_MEMORY_STORE_H

#include <string>
#include <unordered_map>
#include <vector>

#include "gloo_store.h"

namespace horovod {
namespace common {

class MemoryStore : public GlooStore {
public:
  virtual ~MemoryStore()=default;

  void set(const std::string& key, const std::vector<char>& data) override;

  std::vector<char> get(const std::string& key) override;

  void wait(const std::vector<std::string>& keys) override;

  void wait(const std::vector<std::string>& keys,
            const std::chrono::milliseconds& timeout) override;

  void Finalize() override;

private:
  std::unordered_map<std::string, std::vector<char>> map_;
};

} // namespace common
} // namespace horovod

#endif //HOROVOD_GLOO_MEMORY_STORE_H
